/*
 * Visible Human Project Java Program
 * by: Yuh-Jye Chang
 *
 * This Project contains 12 files:
 *
 * VisibleHuman.java  -- the Main Applet
 * ImageFrame.java    -- the Popup Image Frame
 * ImageSelector.java -- Image Object with a Selection Line
 * ImageCanvas.java   -- Image Process and Display Object
 * Scroller.java      -- Motif Look Scrollbar
 * Button3D.java      -- Motif Look 3D Button
 * Panel3D.java       -- Panel with 3D Outlook
 * Label3D.java       -- Label with 3D Outlook
 * Canvas3D.java      -- Canvas with 3D Outlook
 * Thumb.java         -- Thumb in Scrollbar
 * Setting.java       -- Global Setting Interface
 * Delay.java         -- Delay Interrupt Object
 *
 * The Class Hierarchy:
 *
 * Object --- Delay
 * Applet --- VisibleHuman
 * Frame  --- ImageFrame
 * Panel  --- Panel3D --- Scroller
 * Canvas --- Canvas3D -+- Label3D --- Button3D
 *                      |
 *                      +- ImageCanvas --- ImageSelector
 *                      |
 *                      +- Thumb
 *
 * The Component Hierarchy:
 *
 * VisibleHuman -+- board ----+- view[3] --+- logo (ImageCanvas)
 *               |  (Panel3D) |  (Panel)   |  
 *               |            |            +- npac (Button3D)
 *               |            |               
 *               |            +- view[0] --+- select[0] (ImageSelector)
 *               |            |  (Panel)   |  
 *               |            |            +- tagBtn[0] (Button3D)
 *               |            |               ("Axial")
 *               |            +- view[1] --+- select[1] (ImageSelector)
 *               |            |  (Panel)   |  
 *               |            |            +- tagBtn[1] (Button3D)
 *               |            |               ("Sagittal")
 *               |            +- view[2] --+- select[2] (ImageSelector)
 *               |               (Panel)   |  
 *               |                         +- tagBtn[2] (Button3D)
 *               |                            ("Coronal")
 *               +- panel ----+- resolutions -+- tagBtn[3] (Button3D)
 *                  (Panel)   |  (Panel3D)    |  ("Low")
 *                            |               +- tagBtn[4] (Button3D)
 *                            |               |  ("Medium")
 *                            |               +- tagBtn[5] (Button3D)
 *                            |                  ("High")
 *                            +- controls -+- adjust ---+- left (Button3D)
 *                               (Panel3D) |  (Panel3D) |  (triangle)
 *                                         |            +- right (Button3D)
 *                                         |               (triangle)
 *                                         +- slice (Label3D)
 *                                         |  (current slice number)
 *                                         +- load (Button3D)
 *
 * ImageFrame -+- picture ---+- canvas (ImageCanvas)
 *             |  (Panel)    |  (Center)
 *             |             +- hScroll (Scroller)
 *             |             |  (South, HORIZONTAL)
 *             |             +- vScroll (Scroller)
 *             |                (East, VERTICAL)
 *             +- controls --+- Quit (Button3D)
 *                (Panel3D)  |  ("Quit")
 *                           +- Download (Button3D)
 *                           |  ("Download")
 *                           +- Apply (Button3D)
 *                           |  ("Apply")
 *                           +- Org (Button3D)
 *                           |  ("1:1")
 *                           +- lbScale (Panel3D)
 *                           |  (current scale)
 *                           +- sbScale (Scrller)
 *                              (HORIZONTAL)
 *
 * Scroller ----+- up (Button3D)
 * (HORIZONTAL) |  (West, triangle)
 *              +- down (Button3D)
 *              |  (East, triangle)
 *              +- thumb (Thumb)
 *                 (Center)
 *
 * Scroller ----+- up (Button3D)
 * (VERTICAL)   |  (North, triangle)
 *              +- down (Button3D)
 *              |  (South, triangle)
 *              +- thumb (Thumb)
 *                 (Center)
 */

import java.applet.*;
import java.awt.*;
import java.net.*;
import java.io.*;
import java.util.StringTokenizer;

import netscape.security.PrivilegeManager;

import webwisdom.tango.*;

public class VisibleHuman extends Applet implements Runnable, Setting, TDataListener, TControlListener {
	boolean Master = true;
	static ImageFrame frameList = null;
	static URL codeBase, Fatch;
	static double scale = 1.0;
	static int maxp = 2;
	static int step = 50;
	static int wFrame = 580;
	static int hFrame = 420;
	static int offset = 24;
	static int no = 0;
	static int wrap;
	static int sliceOfs = 1001;
	static String LOAD = "Load";
	static String imgdir = "0/";
	static String tagName[] = 
			{"Axial", "Sagittal", "Coronal", 
			 "Low", "Medium", "High"};
	int current = 0, numberofFrames=0;
	int resolution = 0;
	boolean canStop = false;
	Panel panel, view[];
	Panel3D board, resolutions, controls, adjust;
	Label3D slice;
	Button3D tagBtn[], npac, load, left, right;
	ImageCanvas logo;
	ImageSelector select[];
	int sliceNo[] = { 860, 350, 600};

        // TANGO variables
	TAgent agent = null;
	boolean isMaster = true;
	boolean initialized = false;

	// Communication Strings
	final static String	space = " ";
	final static String	initialRequest = "InitialRequest";
	final static String	initialData = "InitialData";
	final static String	updateFrame = "UpdateFrame";
	final static String	quitFrame = "QuitFrame";
	final static String	resizeFrame = "ResizeFrame";
	final static String	originalScale = "OriginalScale";
	final static String	endMessage = "kimd";

	public void init() {
		// TANGO
		try {
			agent = new TAgent(new TLAgentApplet(this));
		}
		catch(TangoException e) {
			e.printStackTrace();
		}

		agent.addTDataListener(this);
		agent.addTControlListener(this);

		isMaster = agent.isMaster();

		int i;
		int defaultRes = 1;
		int defaultView = 2;
		String arg;

		/* Get Current Screen Resolution */
		Dimension screenSize =
			Toolkit.getDefaultToolkit().getScreenSize();

		view = new Panel[4];
		tagBtn = new Button3D[6];
		select = new ImageSelector[3];

		/* Initialize Arguments pass from HTML */
		arg = getParameter("wframe");
		if (arg != null) wFrame = Integer.valueOf(arg).intValue();
		arg = getParameter("hframe");
		if (arg != null) hFrame = Integer.valueOf(arg).intValue();
		arg = getParameter("slicex");
		if (arg != null) sliceNo[0] = Integer.valueOf(arg).intValue();
		arg = getParameter("slicey");
		if (arg != null) sliceNo[1] = Integer.valueOf(arg).intValue();
		arg = getParameter("slicez");
		if (arg != null) sliceNo[2] = Integer.valueOf(arg).intValue();
		arg = getParameter("scale");
		if (arg != null) scale = Double.valueOf(arg).doubleValue();
		arg = getParameter("maxp");
		if (arg != null) maxp = Integer.valueOf(arg).intValue();
		arg = getParameter("step");
		if (arg != null) step = Integer.valueOf(arg).intValue();
		arg = getParameter("defaultRes");
		if (arg != null) defaultRes = Integer.valueOf(arg).intValue();
		arg = getParameter("defaultView");
		if (arg != null) defaultView = Integer.valueOf(arg).intValue();
		arg = getParameter("imgdir");
		if (arg != null) imgdir = arg;
		arg = getParameter("imgbase");

		if (arg != null) {
			try {
				codeBase = new URL(arg);
			}
			catch(Exception e) {
			}
		}
		else {
			try {
				codeBase = new URL("http://kopernik.npac.syr.edu:8080/vishuman/");
			}
			catch(Exception e) {
			}
		}

		/* Set Main Applet to Border Layout */
		setLayout(new BorderLayout());

		/* Setup Components */
		logo = new ImageCanvas();
		logo.setPreferredSize(59, 59);
		logo.mode = 1;

		select[0] = new ImageSelector
			(this, 0, Scrollbar.HORIZONTAL, sliceNo[0], 0, 1728);
		select[0].setValue(sliceNo[0], false);
		select[0].setPreferredSize(100, 59);
		select[0].mode = 1;
		select[1] = new ImageSelector
			(this, 1, Scrollbar.HORIZONTAL, sliceNo[1], 0, 1000);
		select[1].setValue(sliceNo[1], false);
		select[1].setPreferredSize(59, 317);
		select[1].mode = 1;
		select[2] = new ImageSelector
			(this, 2, Scrollbar.VERTICAL, sliceNo[2], 0, 1878);
		select[2].setValue(sliceNo[2], false);
		select[2].setPreferredSize(100, 317);
		select[2].mode = 1;
		npac = new Button3D("NPAC");
		view[3] = new Panel();
		view[3].setLayout(new BorderLayout());
		view[3].add("South", npac);
		view[3].add("Center", logo);
		board = new Panel3D();
		board.add(view[3]);
		for ( i = 0 ; i < 3 ; i++ ) {
			view[i] = new Panel();
			view[i].setLayout(new BorderLayout());
			view[i].add("South", tagBtn[i] =
			new Button3D(Button3D.CHECKBTN, tagName[i]));
			view[i].add("Center", select[i]);
			tagBtn[i].setRatio(2);
			board.add(view[i]);
		}
		resolutions = new Panel3D();
		resolutions.add(tagBtn[3] =
			new Button3D(Button3D.CHECKBTN, tagName[3]));
		resolutions.add(tagBtn[4] =
			new Button3D(Button3D.CHECKBTN, tagName[4]));
		resolutions.add(tagBtn[5] =
			new Button3D(Button3D.CHECKBTN, tagName[5]));
		tagBtn[4].setRatio(3);
		adjust = new Panel3D();
		adjust.setLayout(new FlowLayout(FlowLayout.CENTER, d2, depth));
		adjust.add(left = new Button3D(Button3D.LEFT));
		adjust.add(right = new Button3D(Button3D.RIGHT));
		left.setBase(tagBtn[0].getBase()-8);
		right.setBase(tagBtn[0].getBase()-8);
		controls = new Panel3D();
		controls.add(adjust);
		controls.add(slice = new Label3D(String.valueOf(sliceNo[2])));
		controls.add(load = new Button3D(LOAD));
		load.setRatio(3);
		slice.setActive(false);
		panel = new Panel();
		panel.setLayout(new GridLayout(2, 1));
		panel.add(controls);
		panel.add(resolutions);

		add("Center", board);
		add("South", panel);
		setCurrent(defaultView);
		setResolution(defaultRes+3);
		sliceNo[0] = sliceNo[1] = sliceNo[2] = -1;

		// Caculate offsets for Pupup ImageFrame
		int wrapx = (screenSize.width-offset*4-wFrame)/offset;
		wrap = (screenSize.height-offset*4-hFrame)/offset;
		if ( wrap > wrapx ) wrap = wrapx;
		if ( wrap < 1 ) wrap = 1;
	}

	public void masterChanged(boolean master, String masterName) {
		isMaster = master;
	}

	public void participantJoined(String participantName) {
	}

	public void participantLeft(String participantName) {
	}

	public void audioRequested(boolean isREquested) {
	}

	private byte[] stringToBytes(String s) {
		int l = s.length();
		byte[] t = new byte[l];
		s.getBytes(0, l, t, 0);
		return t;
	}

	private String bytesToString(byte t[]) {
		return new String(t, 0);
	}

	public void destroy() {

		if (agent != null) {
			agent.exit();
		}
		super.destroy();
	}

	void sendMessage(String messageToBeSent) {

		// send the message
		if ( agent != null ) {
			agent.send(stringToBytes(messageToBeSent));
		}
	}

	public void start() {
		// Setup Image into Image Object
		try {
			PrivilegeManager.enablePrivilege("UniversalConnect");
		}
		catch (netscape.security.ForbiddenTargetException e) {
			System.out.println("\tFailed! Permission to connect.");
		}
		catch (Exception e) {
			System.out.println("\tFailed! Unknown exception while enabling privilege.");
		}

		logo.updateImage(getImage(codeBase, "logo.gif"));

		if (isMaster) {
			updatePreview();
			initialized = true;
		}
		else {
			sendMessage(initialRequest + space + endMessage);
		}
	}

	private void updatePreview() {

		try {
			PrivilegeManager.enablePrivilege("UniversalConnect");
		}
		catch (netscape.security.ForbiddenTargetException e) {
			System.out.println("\tFailed! Permission to connect.");
		}
		catch (Exception e) {
			System.out.println("\tFailed! Unknown exception while enabling privilege.");
		}

		/* Update Image only when slice number is changed */
		if ( sliceNo[0] != select[0].getValue()/18 ) {
			sliceNo[0] = select[0].getValue()/18;
			select[1].updateImage(getImage(codeBase, imgdir+
				"JPEGX-18/a_vmX"+(sliceNo[0]+sliceOfs)+"-18.jpg"));
		}
		if ( sliceNo[1] != select[1].getValue()/18 ) {
			sliceNo[1] = select[1].getValue()/18;
			select[2].updateImage(getImage(codeBase, imgdir+
				"JPEGY-18/a_vmY"+(sliceNo[1]+sliceOfs)+"-18.jpg"));
		}
		if ( sliceNo[2] != select[2].getValue()/6 ) {
			sliceNo[2] = select[2].getValue()/6;
			select[0].updateImage(getImage(codeBase, imgdir+
				"JPEG-18/a_vm"+(sliceNo[2]+sliceOfs)+"-18.jpg"));
		}
	}

	public void stop() {
		/* Destroy all the Popup ImageFrame Windows */
		while ( frameList != null ) {
			frameList.dispose();
		}
	}

	private void setCurrent(int _current) {
		/* Select the current Viewport */
		if ( current == _current ) {
			return;
		}
		int i, j, k;
		current = _current;
		k = (current+1)%3;
		showStatus("Select "+tagName[k]+" View");
		for ( i = 0 ; i < 3 ; i++ ) {
			tagBtn[i].setActive(i!=k);
			select[i].setMode((i==k) ? -1 : 1);
			select[i].setColor((i==current) ? Color.cyan : Color.red);
		}
		left.setType((current == 2)? Button3D.UP : Button3D.LEFT);
		right.setType((current == 2)? Button3D.DOWN : Button3D.RIGHT);
		slice.setText(String.valueOf(select[current].getValue()));
		
	}

	private void setResolution(int _resolution) {
		/* Select the current Resolution */
		if ( resolution == _resolution ) return;
		resolution = _resolution;
		showStatus("Set resolution to "+tagName[resolution]);
		for ( int i = 3 ; i < 6 ; i++ ) {
			tagBtn[i].setActive(i!=resolution);
		}
	}

	public void run() {
		/* Fatch the URL and Show Document 
		showStatus(Fatch.toString());
		getAppletContext().showDocument(Fatch);
		*/
	}

	public void showHTML(String newURL) {
		try {
			Fatch = new URL(codeBase, newURL);
			new Delay(100, -1, this);
		}
		catch (MalformedURLException e){
		};

		getAppletContext().showDocument(Fatch,"_black");
	}

	public boolean keyDown(Event ev, int key) {
		if ( key < 0 ) (new Thread(this)).start();
		return true;
	}

	public boolean doAction(String mes) {
		int id;

		if(mes == null) return false;

		try {
			PrivilegeManager.enablePrivilege("UniversalConnect");
		}
		catch (netscape.security.ForbiddenTargetException e) {
			System.out.println("\tFailed! Permission to connect.");
		}
		catch (Exception e) {
			System.out.println("\tFailed! Unknown exception while enabling privilege.");
		}

		StringTokenizer parser = new StringTokenizer(mes);
		if(parser.countTokens()<1) return false;
		String s = parser.nextToken();
		if (s.equals("LOAD")) {
			int x = offset*((no%wrap)+2);
			int slice = select[current].getValue();
			int yScale = 1;
			URL dataBase = codeBase;
			double kScale = 1.0;
			String path = imgdir+"JPEG";
			updatePreview();
			switch ( resolution ) {
				case 3 :
					path += "-L";
					break;
				case 4 :
					path += "-M";
					break;
			}
			switch ( current ) {
				case 0 : // Sagittal Image
					path += "X/a_vmX"+(slice+sliceOfs);
					if ( resolution == 5 ) {
						yScale = 3;
						kScale = scale;
					}
					break;
				case 1 : // Coronal Image
					path += "Y/a_vmY"+(slice+sliceOfs);
					if ( resolution == 5 ) {
						yScale = 3;
						kScale = scale;
					}
					break;
				default : // Axial Image
					path += "/a_vm"+(slice+sliceOfs);
					break;
			}
			switch ( resolution ) {
				case 3 : path += "-L"; break;
				case 4 : path += "-M"; break;
			}
			path += ".jpg";
			//canStop = true;

			Image tmpImage = getImage(dataBase, path);

			no++;
			numberofFrames++;

			new ImageFrame(this,
				"2D Viewer No." + no + " " +
				tagBtn[(current+1)%3].getText()+" "+
				tagBtn[resolution].getText()+" "+
				"Slice#"+slice, path,
				tmpImage, x, x,
				wFrame, hFrame, kScale, 1, yScale, maxp,
				step, no);
			return true;

		}
		else if (s.equals("LEFT")) {
			if(parser.countTokens()<1) return false;
			id = Integer.parseInt(parser.nextToken());
			if ( id != Event.MOUSE_UP ) {
				int v = select[current].getValue()-1;
				if ( select[current].setValue(v, true) ) {
					slice.setText(String.valueOf(v));
				}
			}
			else {
				updatePreview();
			}
			return true;

		}
		else if (s.equals("RIGHT")) {
			if(parser.countTokens()<1) return false;
			id = Integer.parseInt(parser.nextToken());
			if ( id != Event.MOUSE_UP ) {
				int v = select[current].getValue()+1;
				if ( select[current].setValue(v, true) ) {
					slice.setText(String.valueOf(v));
				}
			}
			else {
				updatePreview();
			}
		  return true;
		}
		else if (s.equals("SELECT0")) {
			if(parser.countTokens()<1) return false;
			id = Integer.parseInt(parser.nextToken());
			switch ( id ) {
				case Event.MOUSE_DOWN :
					setCurrent(0);
					break;
				case Event.MOUSE_UP :
					updatePreview();
					break;
				case Event.MOUSE_DRAG :
					break;
				default :
					setCurrent(2);
			}
			slice.setText(String.valueOf(select[current].getValue()));
			return true;
		}
		else if (s.equals("SELECT1")) {
			if(parser.countTokens()<1) return false;
			id = Integer.parseInt(parser.nextToken());  
			switch ( id ) {
				case Event.MOUSE_DOWN :
					setCurrent(1);
					break;
				case Event.MOUSE_UP :
					updatePreview();
					break;
				case Event.MOUSE_DRAG :
					break;
				default :
					setCurrent(0);
			}
			slice.setText(String.valueOf(select[current].getValue()));
			return true;
		}
		else if (s.equals("SELECT2")) {
			if(parser.countTokens()<1) return false;
			id = Integer.parseInt(parser.nextToken());  
			switch ( id ) {
				case Event.MOUSE_DOWN :
					setCurrent(2);
					break;
				case Event.MOUSE_UP :
					updatePreview();
					break;
				case Event.MOUSE_DRAG :
					break;
				default :
					setCurrent(1);
			}
			slice.setText(String.valueOf(select[current].getValue()));
			return true;
		}
		else if (s.equals("TAGBUTTON0")) {
			setCurrent(2);
			return true;
		}
		else if (s.equals("TAGBUTTON1")) {
			setCurrent(0);
			return true;
		}
		else if (s.equals("TAGBUTTON2")) {
			setCurrent(1);
			return true;
		}
		else if (s.equals("TAGBUTTON3")) {
			setResolution(3);
			return true;
		}
		else if (s.equals("TAGBUTTON4")) {
			setResolution(4);
			return true;
		}
		else if (s.equals("TAGBUTTON5")) {
			setResolution(5);
			return true;
		}
		else if (s.equals("SHOWHTML")) {
			showHTML("http://www.npac.syr.edu");
			return true;
		}
		else if (s.equals("SHOWSTRING")) {
			if(parser.countTokens()<1) return false;	
			showHTML(parser.nextToken());
			return true;
		}
		else if (s.equals("LAST")) {
			return true;
		}
	
		return false;
	}

	public void receive (byte b[]) {
		try {
			String mes = bytesToString(b);
			StringTokenizer parser = new StringTokenizer(mes);
			if(parser.countTokens()<1) return;
			String r = parser.nextToken();
			if(r.equals("VH")) {
				doAction(mes.substring(3)); 
			}
			else if(r.equals("IS")) {
				if(parser.countTokens()<1) return;
				r = parser.nextToken();
				if(parser.countTokens()<1) return;
				int id  = Integer.parseInt(parser.nextToken());
				if(r.equals("MOUSEDOWN")) {
					select[id].domouseDown(mes.substring(15));
				}
				else if(r.equals("MOUSEUP")) {
					select[id].domouseUp(mes.substring(13));
					current = -1;
					setCurrent(id);
					updatePreview();
				}
				else if(r.equals("DELAYDRAG")) {
					select[id].dodelayDrag(mes.substring(15));
				}
			}
			else if(r.equals(updateFrame)) {
				if(parser.countTokens()<1) return;

				updateFrameList(mes.substring(updateFrame.length() + 1));
			}
			else if (r.equals(initialRequest)) {
				if ( !isMaster ) {
					return;
				}
				prepareInitialStatus();
			}
			else if (r.equals(initialData)) {
				if ( initialized ) {
					return;
				}
				setInitStatus(mes.substring(initialData.length() + 1));

			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void updateFrameList(String mes) {

		StringTokenizer updateData = new StringTokenizer(mes);

		String updateAction = updateData.nextToken();

		// find the frame to be updated
		ImageFrame foundFrame = frameList;
		int frameID = (Integer.valueOf(updateData.nextToken())).intValue();

		while( foundFrame != null && foundFrame.getID() != frameID ) {
			foundFrame = foundFrame.next;
		}

		if( foundFrame == null ) {
			return;
		}

		if (updateAction.equals(quitFrame)) {
			foundFrame.dispose();
		}
		else if (updateAction.equals(resizeFrame)) {
			double frameScale = (Double.valueOf(updateData.nextToken())).doubleValue();
			double canvasScale = (Double.valueOf(updateData.nextToken())).doubleValue();
			foundFrame.canvas.setScale(frameScale, true);
			foundFrame.Resize.setActive(true);
			foundFrame.Org.setActive(true);
			foundFrame.lbScale.setText(String.valueOf((int)(frameScale*100+0.5))+"%");
			foundFrame.Resize.setActive(true);
    			foundFrame.sbScale.setValue((int)Math.round((Math.log(frameScale)/Math.log(2.0)*step)), true);
		}
		else if (updateAction.equals(originalScale)) {
			double frameScale = foundFrame.getScale();
			double canvasScale = foundFrame.canvas.getScale();

			foundFrame.sbScale.setValue(0, true);

			foundFrame.canvas.setScale(frameScale, true);
			foundFrame.Resize.setActive(true);
			foundFrame.Org.setActive(true);

			foundFrame.lbScale.setText(String.valueOf((int)(frameScale*100+0.5))+"%");
			foundFrame.Resize.setActive(true);
		}

	}

	public void prepareInitialStatus() {
		String str = null;
		int i;

		str =	initialData + space +
			// data for preview
			select[0].getValue() + space +
			select[1].getValue() + space +
			select[2].getValue() + space +
			current + space +
			resolution + space;

		// data for detailed image
		str = str + 
			numberofFrames + space +
			no + space;
		ImageFrame tmpFrame = frameList;
		for ( i = 0 ; i < numberofFrames ; i++ ) {
			// Original data
			str = str +
				tmpFrame.frameID + space +
				tmpFrame.Title + space +
				tmpFrame.path + space +
				tmpFrame.x + space +
				tmpFrame.y + space +
				tmpFrame.w + space +
				tmpFrame.h + space +
				tmpFrame.scale + space +
				tmpFrame.xScale + space +
				tmpFrame.yScale + space +
				tmpFrame.maxp + space +
				tmpFrame.step + space +

			// updated data (if any)
				tmpFrame.getScale() + space +
				tmpFrame.canvas.getScale() + space;

			tmpFrame = tmpFrame.next;
		}

		str = str + endMessage;

		sendMessage(str);
	}

	public void setInitStatus(String mes) {
		int i;

		try {
			PrivilegeManager.enablePrivilege("UniversalConnect");
		}
		catch (netscape.security.ForbiddenTargetException e) {
			System.out.println("\tFailed! Permission to connect.");
		}
		catch (Exception e) {
			System.out.println("\tFailed! Unknown exception while enabling privilege.");
		}

		StringTokenizer initData = new StringTokenizer(mes);

		// data for preview
		int imageNo0 = (Integer.valueOf(initData.nextToken())).intValue();
		int imageNo1 = (Integer.valueOf(initData.nextToken())).intValue();
		int imageNo2 = (Integer.valueOf(initData.nextToken())).intValue();

		int mastercurrent = (Integer.valueOf(initData.nextToken())).intValue();
		int masterresolution = (Integer.valueOf(initData.nextToken())).intValue();

		select[0].setValue(imageNo0, true);
		select[1].setValue(imageNo1, true);
		select[2].setValue(imageNo2, true);
		sliceNo[0] = -1;
		sliceNo[1] = -1;
		sliceNo[2] = -1;
		updatePreview();

		slice.setText(String.valueOf(select[mastercurrent].getValue()));
		setCurrent(mastercurrent);
		setResolution(masterresolution);
		// data for detailed image
		numberofFrames = (Integer.valueOf(initData.nextToken())).intValue();
		no = (Integer.valueOf(initData.nextToken())).intValue();

		for ( i = 0 ; i < numberofFrames ; i++ ) {

			// data for preview
			int frameID = (Integer.valueOf(initData.nextToken())).intValue();

			int j;
			String Title = null;
			for ( j = 0 ; j < 6 ; j++ ) {
				Title = Title + initData.nextToken() + space;
			}

			String path = initData.nextToken();
			int x = (Integer.valueOf(initData.nextToken())).intValue();
			int y = (Integer.valueOf(initData.nextToken())).intValue();
			int w = (Integer.valueOf(initData.nextToken())).intValue();
			int h = (Integer.valueOf(initData.nextToken())).intValue();
			double scale = (Double.valueOf(initData.nextToken())).doubleValue();
			int xScale = (Integer.valueOf(initData.nextToken())).intValue();
			int yScale = (Integer.valueOf(initData.nextToken())).intValue();
			int maxp = (Integer.valueOf(initData.nextToken())).intValue();
			int step = (Integer.valueOf(initData.nextToken())).intValue();

			// updated data (if any)
			double frameScale = (Double.valueOf(initData.nextToken())).doubleValue();
			double canvasScale = (Double.valueOf(initData.nextToken())).doubleValue();

			ImageFrame tmpFrame = new ImageFrame(
						this, Title, path,
						getImage(codeBase, path),
						x, y, w, h,
						scale, xScale, yScale,
						maxp, step, frameID);

			tmpFrame.canvas.setScale(frameScale, true);
			tmpFrame.Org.setActive(canvasScale != 1.0);
			tmpFrame.Org.setActive(true);
			tmpFrame.lbScale.setText(String.valueOf((int)(frameScale*100+0.5))+"%");
			tmpFrame.Resize.setActive(true);

    			tmpFrame.sbScale.setValue((int)Math.round((Math.log(frameScale)/Math.log(2.0)*step)), true);

		}

		initialized = true;

	}

	public boolean action(Event ev, Object arg) {
		String s = endMessage;
		boolean sending = false;

		if (arg == select[0]) {
			s = "SELECT0 "+ev.id;
		} else if (arg == select[1]) {
			s = "SELECT1 "+ev.id;
		} else if (arg == select[2]) {
			s = "SELECT2 "+ev.id;
		}

		if ( s.equals(endMessage) ) {
			sending = true;
		}
		else {
			sending = false;
		}

		if ( arg == load ) {
			s = "LOAD";
		} else if (arg == left) {
			s = "LEFT "+ev.id;
		} else if (arg == right) {
			s = "RIGHT "+ev.id;
		} else if (arg == tagBtn[0]) {
			s = "TAGBUTTON0";
		} else if (arg == tagBtn[1]) {
			s = "TAGBUTTON1";
		} else if (arg == tagBtn[2]) {
			s = "TAGBUTTON2";
		} else if (arg == tagBtn[3]) {
			s = "TAGBUTTON3";
		} else if (arg == tagBtn[4]) {
			s = "TAGBUTTON4";
		} else if (arg == tagBtn[5]) {
			s = "TAGBUTTON5";
		} else if ((arg == logo) || (arg == npac)) {
			if ( (ev.id == Event.MOUSE_DOWN) || (arg == npac) ) {
				s = "SHOWHTML";
			}
		} else if (arg instanceof String) {
			s = "SHOWSTRING "+ (String)arg;
		}

		if ( s.equals(endMessage) ) {
			return false;
		}

		if ( sending ) {
			String ss = "VH "+ s;
			sendMessage(ss);
		}

		if (doAction(s)) {
			return true;
		}

		return false;
	}
}
